
#include <FSTREAM.H>    //also includes iostream.h
#include <IOMANIP.H>	//for setw()
#include <PROCESS.H>	//for exit()
#include <CONIO.H>      //for getche(), etc.
#include <STRING.H>		//for strcpy(), etc.
#include <DIR.H>		//for changing extension - fnsplit()
#include <CTYPE.H>		//for to_upper()
#include "COMPEXP.H"

#define COMP_DEFAULT_EXT ".CMP"
#define COMP_SECOND_EXT ".CPR"
#define EXP_DEFAULT_EXT ".EXP"
#define EXP_SECOND_EXT ".EPN"
const unsigned int MAX_FILENAME_LEN = 40;

enum boolean { false, true };
enum modetype { compress, expand };

modetype mode;	//global variable: compress / expand

int main(int argc, char* argv[])
	{
	//data
	char src_file[MAX_FILENAME_LEN] = "";
	char dest_file[MAX_FILENAME_LEN] = "";
	char* code_strings[256];	//new binary codes for characters

	//function declarations
	void ask_src_filename(char*);
	void read_arguments(int, char**, char*, char*);
	void change_extension(char*, char*);
	void write_new_code_to_file(char**, char*, char*);
	int to_expand(char*, char*, unsigned long, node_tree&);
	unsigned long file_length(char*);

	char ch;
	do
		{
		cout << "\nCompress/Expand? ";
		ch = toupper(getche());
		}
	while ((ch!='C') && (ch!='E'));
	mode = (ch == 'C')?  compress : expand;
	cout << endl;

	//get filenames from user
	read_arguments(argc, argv, src_file, dest_file);
	if (src_file[0] == 0)
		ask_src_filename(src_file);
	if (dest_file[0] == 0)
		change_extension(dest_file, src_file);

	node_tree tree;
	if (mode==compress)
		{
		cout << " Compressing " << src_file
			 << " into " << dest_file << " . . .";

		tree.read_from_file(src_file);
		tree.do_Hauffman();				//brains of program
		tree.form_codes(code_strings);
		tree.write_to_file(src_file, dest_file);
		write_new_code_to_file(code_strings, src_file, dest_file);

		unsigned long int old_size = file_length(src_file);
		unsigned long int new_size = file_length(dest_file);
		cout << "completed."
			 << "\n Old Length: " << old_size
			 << "\n New Length: " << new_size
			 << "\n Compression Percentage: " <<(100-(signed int)((new_size*100) / old_size))
			 << "%" << endl;

		return 0;
		}
	else
		{
		cout << "Expanding " << src_file
		 << " to " << dest_file << "...";

		unsigned long int start_pos = tree.read_tree_from_file(src_file);
		int result = to_expand(src_file, dest_file, start_pos, tree);

		if (result==1) cout << "\nFile shorter than required." << endl;
		if (result==2) cout << "\nFile longer than expected." << endl;
		if (result != 0)
			{
			cout << "Maybe the file is damaged, or the file has not been "
				"compressed using compress.exe!" << endl;
			return result;
			}

		cout << "completed.";
		return 0;
		}
	}

void read_arguments(int argc, char** argv, char* filename1,char* filename2)
	{
	boolean is_file_valid(char*);
	boolean test_write_to_file(char*);
	void to_upper_str(char*);
	unsigned long file_length(char*);

	if (argc == 1) return;

	if (argc > 3)
		cout << "\nThese many arguments unexpected!" << endl
			 << "Syntax:    compexp <source file> [<dest file>] " << endl
			 << " (If dest file not specified, source file with default"
			 << " extension will be chosen.)" << endl;

	to_upper_str(argv[1]);

	if (!is_file_valid(argv[1]))
		{
		cout << "\nArgument: "<< argv[1]
			 << " is not a valid source file!" << endl;
		return;
		}

	if (file_length(argv[1]))
		strcpy(filename1, argv[1]);
	else
		{
		cout << "Source file " << argv[1] << "is of zero length!\n";
		return;
		}

	if (argc == 2) return;

	to_upper_str(argv[2]);

	if (!strcmp(filename1, argv[2]))
		{
		cout << "Source & Target files cannot be same!!!\n";
		return;
		}

	if (is_file_valid(argv[2]))
		{
		cout <<"The destination file "<< argv[2] << " already exists!";
		char ch;
		do
			{
			cout << "\nOverwrite(Y/N)?";
			ch = getche();
			}
		while ((ch!='y') && (ch!='Y') && (ch!='n') && (ch!='N'));
		cout << "\n";

		if ((ch=='n') || (ch=='N'))
			return;
		}

	if (!test_write_to_file(argv[2]))
		{
		cout << "\nError on opening file "<< argv[2] << "\n";
		return;
		}

	strcpy(filename2, argv[2]);
	}

void ask_src_filename(char* filename)
	{
	boolean is_file_valid(char*);
	void to_upper_str(char*);

	boolean done;
	do
		{
		if (mode==compress)
			cout << "\nEnter the name of file to compress: ";
		else
			cout << "\nEnter the name of file to expand: ";
		cin >> setw(MAX_FILENAME_LEN) >> filename;
		//cin.get(filename, MAX_FILENAME_LEN) gives erroneous results

		if (!is_file_valid(filename))
			{
			cout << "File invalid.";

			char ch;
			do
				{
				cout << "\nAbort/Retry?";
				ch = getche();
				}
			while ( (ch!='A') && (ch!='a') && (ch!='r') && (ch!='R') );

			if ((ch=='a') || (ch=='A'))
				exit(1);
			else
				done = false;
			}
		else
			done = true;
		}
	while (!done);

	 to_upper_str(filename);
	}

void change_extension(char* filename1, char* filename2)
	{
	char drive[MAXDRIVE], dir[MAXDIR];
	char name[MAXFILE], ext[MAXEXT] = "";

	fnsplit(filename2, drive, dir, name, ext);

	strcpy(filename1, name);
	if (mode==compress)
		{
		if (!strcmp(ext, COMP_DEFAULT_EXT))   //strcmp returns 0 if equal
			strcat(filename1, COMP_SECOND_EXT);
		else
			strcat(filename1, COMP_DEFAULT_EXT);
		}
	else
		{
		if (!strcmp(ext, EXP_DEFAULT_EXT))   //strcmp returns 0 if equal
			strcat(filename1, EXP_SECOND_EXT);
		else
			strcat(filename1, EXP_DEFAULT_EXT);
		}
	}

boolean is_file_valid(char* filename)
	{
	boolean ret_val;

	ifstream in_file;
	in_file.open(filename, ios::in | ios::binary | ios::nocreate);

	if (in_file.rdstate())
		ret_val = false;
	else
		ret_val = true;

	in_file.close();

	return ret_val;
	}

boolean test_write_to_file(char* filename)
	{
	boolean ret_val;

	ofstream out_file;
	out_file.open(filename, ios::out | ios::binary);

	if (out_file.rdstate())
		ret_val = false;
	else
		ret_val = true;

	out_file.close();

	return ret_val;
	}

void write_new_code_to_file(char** codes, char* src_file, char* dest_file)
	{
	void bits_to_file(char*, int=0);

  //open source and destination files
	ifstream in_file; in_file.open(src_file, ios::in | ios::binary);
	bits_to_file(dest_file, 1);	//tell bits_to_file the destination file name

	unsigned char ch;
	for( ; ; )		//infinite loop
		{
		in_file.get(ch);
		if (!in_file) break; //because, AFTER (-1) has been read
							 // the in_file object becomes zero

		bits_to_file(codes[ch]);
		}

  //close source and destination files
	bits_to_file(0, 2); //tell bits_to_file to close file
	in_file.close();
	}

void bits_to_file(char* bitpatt, int flag)
	{
	unsigned char bin_to_dec (char*);

	static char string[300] = "";
	static ofstream out_file;

	if (flag==1)
		{ 				//First Time
		out_file.open(bitpatt, ios::binary | ios::app);
		return;  //this time bitpatt actually contains dest_filename
		}

	if (flag==2)
		{                  //Last Time
		//store remaining bits in string
		if (strlen(string))
			{
			char temp[] = "00000000";    // insert zeroes so as to make it
			strcpy(temp, string);        // 8 bits, these xtra bits do not
			temp[strlen(string)] = '0';  // matter,bcos filelength is also
			out_file.put(bin_to_dec(temp)); //being stored.
			}

		out_file.close();
		return;
		}

	 strcat(string, bitpatt);

	while (strlen(string) >= 8)
		{ 	//ready to be written to file
		out_file.put(bin_to_dec(string));

		for (int i=0; string[i+7]; i++)
			string[i] = string[i+8];
		}
	}

unsigned char bin_to_dec(char* pattern)
	{
	unsigned char num = 0;

	for (int i=0; i<=7; i++)
		num = (num << 1) + (pattern[i] - '0');

	return num;
	}

unsigned long file_length(char* filename)
	{
	unsigned long int file_size;
	ifstream temp(filename);

	temp.seekg(0L, ios::end);
	file_size = temp.tellg();

	temp.close();

	return file_size;
	}

void to_upper_str(char* string)
	{
	for (int i=0; i< strlen(string); i++)
		string[i] = toupper(string[i]);
	}

int to_expand(char* src_file, char* dest_file,
								unsigned long start_pos, node_tree& tree)
	{
	unsigned long int file_length(char*);	//proto

	int to_return = 0;  //success
	unsigned long int file_size = file_length(src_file);

	if (!start_pos)  //message from node_tree::read_from_file
		return 1;   //that file was shorter than required

	ifstream in_file;
	in_file.open(src_file, ios::in | ios::binary);
	in_file.seekg(start_pos);

	ofstream out_file;
	out_file.open(dest_file, ios::out | ios::binary);

	unsigned char key = 1;
	long int rest;
	int dirctn, retval;
	unsigned char ch;
	while (1)	//infinite loop
		{
		in_file.get(ch);

		if (!in_file)
			{    //file should not end here
			to_return = 1;
			goto all_read;
			}

			for (key=128; key; key>>=1)
			{
			dirctn = (ch & key) ?  1 : 0;

			retval = tree.go(dirctn);

			if (retval != 1000)
				{
				out_file.put (retval % 256);
				if (retval > 255)
					{
					rest = file_size - in_file.tellg();
					if (rest) to_return = 2;   //file was too long
					goto all_read;
					}
				}
			}
		}
	all_read:
	in_file.close();
	out_file.close();

	return to_return;
	}
